#include <limits.h>
#include <time.h>
#include "msg.h"
#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "cache.h"

#define st_mtime st_mtim.tv_sec

#ifdef __APPLE__
         #define st_mtime st_mtimespec.tv_sec
#endif
// TODO: maybe define for fopen things

void
cleaning(unsigned int hits_ref) /* delete every directory containing hits
                                   below hits_ref */
{
    DIR *dir_cache = opendir(".");
    struct dirent *dir = NULL;
    char path[MAX_LEN_DNS + 8];

    info("cleaning");

    if(dir_cache != NULL)
    {
        while((dir = readdir(dir_cache)) != NULL)
        {
            if(strncmp(dir->d_name,".",1) == 0 || strncmp(dir->d_name,"..",2) == 0)
                continue;

            if(get_hits(dir->d_name) <= hits_ref)
            {
                mkpath(path,dir->d_name,HITS);
                if(unlink(path) == -1) error("cleaning() rm hits");
                mkpath(path,dir->d_name,ANSWER);
                if(unlink(path) == -1) error("cleaning() rm answer");
                if(rmdir(dir->d_name) == -1) error("cleaning() rmdir");
            }
        }
        closedir(dir_cache);
    }
    else error("cleaning() opendir");
}

void
find_min_hits(unsigned int *hits,char *dns) /* find dns whith the lowest hits */
{
    DIR *d;
    struct dirent *dir;
    struct stat st;
    unsigned int hits_min = UINT_MAX;
    unsigned int h = 0;
    time_t oldest = 0;
    char path[MAX_LEN_DNS + 8];

    time(&oldest);
    d = opendir(".");

    if (d != NULL)
    {
        while ((dir = readdir(d)) != NULL)
        {
            if(strncmp(dir->d_name,".",1) == 0 || strncmp(dir->d_name,"..",2) == 0)
                continue;

            h = get_hits(dir->d_name);

            if(h < hits_min)
            {
                mkpath(path,dir->d_name,HITS);
                stat(path,&st);

                if(st.st_mtime < oldest)
                {
                    oldest = st.st_mtime;
                    hits_min = h;
                    dns = strncpy(dns,dir->d_name,MAX_LEN_DNS);
                }
            }
        }
        closedir(d);
    }
    else
    {
        hits_min = 0;
        error("find_min_hits() opendir");
    }

    *hits = hits_min;
}

unsigned int
get_hits(char *key)
{
    unsigned int hits = -1;
    FILE *tmp = NULL;
    struct stat st;
    char path[MAX_LEN_DNS + 8];

    mkpath(path,key,HITS);

    if(path != NULL)
    {
        if(stat(path,&st) != 0)
        {
            error("get_hits() hits file doesn't exists... recreate");
            hits = 1;
            set_hits(path,hits);
        }
        else
        {
            tmp = fopen(path,"rb");

            if(tmp != NULL)
            {
                fread(&hits,sizeof(unsigned int),1,tmp);
                fclose(tmp);
            }
            else error("get_hits() opening hits");
        }
    }

    return hits;
}

void
set_hits(char *key,unsigned int hits)
{
    FILE *tmp = NULL;
    char path[MAX_LEN_DNS + 8];

    mkpath(path,key,HITS);

    if(path != NULL)
    {
        tmp = fopen(path,"wb");

        if(tmp != NULL)
        {
            fwrite(&hits,sizeof(unsigned int),1,tmp);
            fclose(tmp);
            if(chmod(path,S_IRUSR|S_IWUSR) == -1) error("set_hits() chmod");
        }
        else error("set_hits() opening");
    }
    else error("set_hits() path is NULL");
}

unsigned int
inc_hits(char *key)
{
    FILE *hits_file = NULL;
    unsigned int hits = 0;
    struct stat st;
    char path[MAX_LEN_DNS + 8];

    mkpath(path,key,HITS);

    if(stat(path,&st) != 0)
    {
        error("inc_hits() hits file should exists... recreate");
        hits = 1;
        set_hits(path,hits);
    }
    else
    {
        hits_file = fopen(path,"r+b");
        if(hits_file != NULL)
        {
            fread(&hits,sizeof(unsigned int),1,hits_file);
            hits++;
            fseek(hits_file,0,SEEK_SET);
            fwrite(&hits,sizeof(unsigned int),1,hits_file);
            fclose(hits_file);
        }
        else error("inc_hits() opening");
    }

    return hits;
}

void
set_answer(char *key,unsigned char *buf,int len_answer)
{
    FILE *ans = NULL;
    char path[MAX_LEN_DNS + 8];

    mkpath(path,key,ANSWER);

    ans = fopen(path,"wb");

    if(ans != NULL)
    {
        fwrite(&buf[2],len_answer - 2,1,ans);
        fclose(ans);
        if(chmod(path,S_IRUSR) == -1) error("set_answer() chmod");
    }
    else error("set_answer() opening");
}

int
get_answer(char *key,unsigned char *buf,unsigned char *id)
{
    FILE *tmp = NULL;
    int len_answer = 0;
    char path[MAX_LEN_DNS + 8];

    mkpath(path,key,ANSWER);

    tmp = fopen(path,"rb");
    if(tmp != NULL)
    {
        buf[0] = id[0]; buf[1] = id[1];
        len_answer = fread(&buf[2],1,LEN_BUF,tmp);
        fclose(tmp);
    }
    else error("get_answer()");
    return feof(tmp) == 0 ? 0 : len_answer;
}

char
*mkpath(char *path,char *root,char *file)
{
    memset(path,MAX_LEN_DNS + 8,0);
    snprintf(path,MAX_LEN_DNS + 8,"%s/%s",root,file);

    return path;
}

int
count_cache(void)
{

    DIR *dir_cache = opendir(".");
    struct dirent *dir = NULL;
    int i = 0;

    if(dir_cache != NULL)
    {
        while((dir = readdir(dir_cache)) != NULL)
        {
            if(strncmp(dir->d_name,".",1) == 0 || strncmp(dir->d_name,"..",2) == 0)
                continue;
            i++;
        }
    }
    else error("count_cache() opendir");

    return i;
}
