#include "config.h"
#include "cache.h"
#include "msg.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>

#define _XOPEN_SOURCE 700
#define NB_SERVER sizeof(dns_servers) / sizeof(dns_servers[0])


int extract_dns(unsigned char* buf,char *dns);
int resolve(int s_dns,unsigned char *buf,int len,struct sockaddr_in dns,int s_proxy,
        struct sockaddr_in proxy);
int check_valid_dns_name(char *name);

/*
   errors handling : when we encounter an error when opening or reading of file
   we don't exit but continue, otherwise the system may not have any resolver. If an
   answer file doesn't exists but the directory dns does we should redo the query
   to the dns and store it again idem if we wanna increments the hit.
*/


int main()
{
    struct stat st;
    unsigned char buf[LEN_BUF];
    char name_query[MAX_LEN_DNS];
    char dns_min[MAX_LEN_DNS];
    struct sockaddr_in proxy;
    struct sockaddr_in dns;
    socklen_t len_sock;
    int len_answer,len_proxy;
    int dns_in_cache = 0;
    int s_proxy;
    int s_dns;
    unsigned char id[2];
    unsigned int hits_min = 0; /* unsigned to restart at 0 and do not owerflow*/
    unsigned char count = 0; /* same reason but we don't need integer size to
                                cycle the dns*/

    if(stat(PATH_CACHE,&st) != 0)
        fatal(mkdir(PATH_CACHE,S_IRWXU),"creating cache dir");

    fatal(chdir(PATH_CACHE),"entering cache dir");

    s_proxy = socket(AF_INET , SOCK_DGRAM , IPPROTO_UDP);
    s_dns = socket(AF_INET , SOCK_DGRAM , IPPROTO_UDP);

    fatal(s_proxy,"error creating socket proxy");
    fatal(s_dns,"error creating socket dns");

    dns_in_cache = count_cache();
    find_min_hits(&hits_min,dns_min);

    memset(&proxy,0,sizeof(proxy));
    memset(&dns,0,sizeof(dns));

    proxy.sin_family = AF_INET;
    proxy.sin_port = htons(53);
    proxy.sin_addr.s_addr = htonl(INADDR_ANY);

    dns.sin_family = AF_INET;
    dns.sin_port = htons(53);
    dns.sin_addr.s_addr = inet_addr(dns_servers[0]);
    len_sock = sizeof(dns);

    fatal(bind(s_proxy,(struct sockaddr *)&proxy,sizeof(proxy)), "error binding");

    len_proxy = recvfrom(s_proxy,buf,LEN_BUF,0,(struct sockaddr*)&proxy,&len_sock);
    while(len_proxy != -1)
    {
        if(extract_dns(buf,name_query) == -1)
            continue;

        info(name_query);

        if(stat(name_query,&st) == 0)
        {
            info("dns is in cache exists");

            id[0] = buf[0]; id[1] = buf[1];
            len_answer = get_answer(name_query,buf,id);

            if(len_answer == 0)
            {
                warning("answer can't be read... recreate");
                dns.sin_addr.s_addr = inet_addr(dns_servers[count++ % NB_SERVER]);
                len_answer = resolve(s_dns,buf,len_proxy,dns,s_proxy,proxy);
                set_answer(name_query,buf,len_answer);
            }
            else
            {
                info("answer is in cache");
                len_proxy = sendto(s_proxy,buf,len_answer + 2,0,(struct sockaddr *)&proxy,len_sock);
                fatal(len_proxy,"sendto proxy failed");
            }

            inc_hits(name_query);

            if(strncmp(dns_min,name_query,MAX_LEN_DNS) == 0)
                find_min_hits(&hits_min,dns_min);
        }
        else
        {
            info("dns not in cache. creating");
            dns.sin_addr.s_addr = inet_addr(dns_servers[count++ % NB_SERVER]);
            len_answer = resolve(s_dns,buf,len_proxy,dns,s_proxy,proxy);

            if(mkdir(name_query,S_IRWXU) == 0)
            {
                set_answer(name_query,buf,len_answer);
                set_hits(name_query,1);
                dns_in_cache++;
            }
            else error("can't create dns directory");
        }

        if(dns_in_cache >= NB_DNS - 1)
            cleaning(hits_min);

        len_proxy = recvfrom(s_proxy,buf,LEN_BUF,0,(struct sockaddr*)&proxy,&len_sock);
    }

    close(s_dns);
    close(s_proxy);
    fatal(len_answer,"recvfrom proxy failed");

    return -1;
}

int
resolve(int s_dns,unsigned char *buf,int len,struct sockaddr_in dns,int s_proxy,
        struct sockaddr_in proxy)
{
    socklen_t len_sock;

    len = sendto(s_dns,buf,len,0,(struct sockaddr*)&dns,sizeof(dns));
    fatal(len,"sendto() dns resolve failed");

    len = recvfrom (s_dns,buf,LEN_BUF,0,(struct sockaddr*)&dns,&len_sock);
    fatal(len,"recvfrom() dns failed");

    if(buf[2] & 1 << 7)
    {
        len = sendto(s_proxy,buf,len,0,(struct sockaddr*)&proxy,len_sock);
        fatal(len,"sendto() proxy failed");
    }
    else error("didn't receive a response from dns");

    return len;
}

int /* dns is supposed to be an array of 255 char */
extract_dns(unsigned char* buf,char *dns)
{
    int i = 12; /* start of 03abc06abcdef03com in the buf answer*/
    int next_dot;
    int len_dns = -1;

    for(i = 12;buf[i] != '\0';i += buf[i] + 1)
        len_dns += buf[i] + 1;

    if(len_dns > MAX_LEN_DNS)
    {
        warning("dns too long");
        return -1;
    }

    dns = memcpy(dns,&buf[13],len_dns + 1);
    dns[len_dns + 1] = '\0';

    i = buf[12];

    while(dns[i] != '\0')
    {
        next_dot = i + dns[i] + 1;
        dns[i] = '.';
        i = next_dot;
    }

    return check_valid_dns_name(dns);
}

int
check_valid_dns_name(char *name)
{
    int i = 0;
    int count = 0;
    char c = 0;

    /* every label must be < 63
       the total length must be < 255 (checked in extract_dns)
       only [.-a-Z0-9] character for label
     */

    for(i = 0;i < strlen(name);i++)
    {
        c = name[i];

        if(!isupper(c) && !islower(c) && !isdigit(c) && c != '.' && c != '-')
            return -1;

        if(c == '.')
        {
            if(count > 63 || count == 0)
            {
                warning("invalid length for label");
                return -1;
            }
            count = 0;
        }
        else count++;
    }

    if(count > 63)
    {
        warning("label is over 63");
        return -1;
    }
    return 0;
}
